/*
 * Copyright 2018 Jérôme Glisse
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <GL/glew.h>
#include <GL/freeglut.h>

struct vid {
    GLuint vao;
    GLuint text;
    GLuint uvbo;
    GLuint prog;
    GLuint posbo;
    GLuint vshader;
    GLuint fshader;
    GLuint pboidx;
    GLuint pbos[2];
    unsigned width;
    unsigned height;
    unsigned posx;
    unsigned posy;
    unsigned size;
    uint32_t *ptr;
};

struct packet {
    struct packet *next;
    unsigned vidid;
    unsigned x;
    unsigned y;
    unsigned w;
    unsigned h;
    uint32_t *ptr;
};

struct net {
    struct packet *packet;
    uint8_t *buffer;
    unsigned size;
    unsigned offset;
};

#define MAX_VIDS 16

struct context {
    struct net net;
    unsigned win_width;
    unsigned win_height;
    unsigned vid_count;
    struct vid vids[MAX_VIDS];
};

struct context ctx = {0};

void check_compile(GLuint id, const char *name)
{
    char msg[64 << 10];
    GLint status;

    glGetShaderiv(id, GL_COMPILE_STATUS, &status);
    if (!status) {
        glGetInfoLogARB(id, sizeof(msg), NULL, msg);
        printf("%s: %s\n", name, msg);
        exit(1);
    }
}

void check_link(GLuint id, const char *name)
{
    char msg[64 << 10];
    GLint status;

    glGetProgramiv(id, GL_LINK_STATUS, &status);
    if (!status) {
        glGetInfoLogARB(id, sizeof(msg), NULL, msg);
        printf("%s: %s\n", name, msg);
        exit(1);
    }
}

static void vid_init(struct vid *vid,
                     unsigned w, unsigned h,
                     unsigned x, unsigned y)
{
    float uv[8] = {1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 0.0f, 0.0f, 0.0f};
    const char* vshader =
        "#version 400\n"
        "in vec2 vert_uv;"
        "in vec3 vert_pos;"
        "out vec2 frag_uv;"
        "void main() {"
        "    frag_uv = vert_uv;"
        "    gl_Position = vec4(vert_pos, 1.0);"
        "}";
    const char* fshader =
        "#version 400\n"
        "in vec2 frag_uv;"
        "out vec4 color;"
        "uniform sampler2D tex;"
        "void main() {"
        "    color = texture(tex, frag_uv);"
        "}";
    uint32_t *tmp;

    vid->width = w;
    vid->height = h;
    vid->posx = x;
    vid->posy = y;
    vid->size = w * h * 4;

    vid->vshader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vid->vshader, 1, &vshader, NULL);
    glCompileShader(vid->vshader);
    check_compile(vid->vshader, "vert");
    vid->fshader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(vid->fshader, 1, &fshader, NULL);
    glCompileShader(vid->fshader);
    check_compile(vid->fshader, "frag");
    vid->prog = glCreateProgram();
    glAttachShader(vid->prog, vid->fshader);
    glAttachShader(vid->prog, vid->vshader);
    glBindAttribLocation(vid->prog, 0, "vert_uv");
    glBindAttribLocation(vid->prog, 1, "vert_pos");
    glLinkProgram(vid->prog);
    check_link(vid->prog, "prog");

    glGenBuffers(1, &vid->uvbo);
    glBindBuffer(GL_ARRAY_BUFFER, vid->uvbo);
    glBufferData(GL_ARRAY_BUFFER, sizeof(uv), uv, GL_STATIC_DRAW);
    glGenBuffers(1, &vid->posbo);
    glBindBuffer(GL_ARRAY_BUFFER, vid->posbo);
    glGenVertexArrays(1, &vid->vao);
    glBindVertexArray(vid->vao);
    glBindBuffer(GL_ARRAY_BUFFER, vid->uvbo);
    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, NULL);
    glBindBuffer(GL_ARRAY_BUFFER, vid->posbo);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, NULL);

    glGenTextures(1, &vid->text);
    glBindTexture(GL_TEXTURE_2D, vid->text);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
    tmp = malloc(vid->size);
    memset(tmp, 128, vid->size);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, vid->width, vid->height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, tmp);
    glBindTexture(GL_TEXTURE_2D, 0);
    free(tmp);

    vid->pboidx = 0;
    glGenBuffers(2, vid->pbos);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, vid->pbos[0]);
    glBufferData(GL_PIXEL_UNPACK_BUFFER, vid->size, 0, GL_STREAM_DRAW_ARB);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, vid->pbos[1]);
    glBufferData(GL_PIXEL_UNPACK_BUFFER, vid->size, 0, GL_STREAM_DRAW_ARB);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
}

static void vid_draw(struct vid *vid)
{
    float verts[12];
    float px, py, w, h;

    w = (float)vid->width / ctx.win_width;
    h = (float)vid->height / ctx.win_height;
    px = ((2.0f * vid->posx) / ctx.win_width) - 1.0f;
    py = ((-2.0f * vid->posy) / ctx.win_height) + 1.0f;
    verts[ 0] =  px + w; verts[ 1] =  py - 0; verts[ 2] = 0.0f;
    verts[ 3] =  px + 0; verts[ 4] =  py - 0; verts[ 5] = 0.0f;
    verts[ 6] =  px + w; verts[ 7] =  py - h; verts[ 8] = 0.0f;
    verts[ 9] =  px + 0; verts[10] =  py - h; verts[11] = 0.0f;

    glBindBuffer(GL_ARRAY_BUFFER, vid->posbo);
    glBufferData(GL_ARRAY_BUFFER, sizeof(verts), verts, GL_STATIC_DRAW);

    glUseProgram(vid->prog);
    glBindVertexArray(vid->vao);
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, vid->text);
    glUniform1i(glGetUniformLocation(vid->prog, "tex"), 0);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

static void vid_upload_start(struct vid *vid)
{
    glBindTexture(GL_TEXTURE_2D, vid->text);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, vid->pbos[vid->pboidx]);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, vid->width, vid->height,
                    GL_RGBA, GL_UNSIGNED_BYTE, NULL);

    vid->pboidx = (vid->pboidx + 1) & 1;
}

static void vid_upload_stop(struct vid *vid)
{
    if (vid->ptr) {
        glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);
        vid->ptr = NULL;
    }
}

static void vid_upload_packet(struct vid *vid, struct packet *pkt)
{
    unsigned x, y, vpitch, pitch, mx, my, a, b;

    if(pkt->x >= vid->width || pkt->y >= vid->height) {
        return;
    }

    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, vid->pbos[vid->pboidx]);
    glBufferData(GL_PIXEL_UNPACK_BUFFER, vid->size, 0, GL_STREAM_DRAW);
    vid->ptr = (uint32_t *)glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);

    if(vid->ptr == NULL || pkt->x >= vid->width || pkt->y >= vid->height) {
        return;
    }
    vpitch = vid->width;
    pitch = pkt->w;
    mx = pkt->x + pkt->w;
    mx = mx > vid->width ? vid->width : mx;
    my = pkt->y + pkt->h;
    my = my > vid->height ? vid->height : my;

#if 1
    for (y = pkt->y, b = 0; y < my; ++y, ++b) {
#if 0
        for (x = pkt->x, a = 0; x < mx; ++x, ++a) {
            vid->ptr[y * vpitch + x] = pkt->ptr[b * pitch + a];
        }
#else
        memcpy(&vid->ptr[y * vpitch + pkt->x], &pkt->ptr[b * pitch], pitch * 4);
#endif
    }
#endif

    glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);
    vid->ptr = NULL;
}

static void network_init(struct net *net, unsigned size)
{
    net->offset = 0;
    net->size = size;
    net->packet = NULL;
    net->buffer = malloc(size);
}

static void network_process_packet(struct net *net)
{
    float stime, etime, dtime;
    unsigned long bytes = 0, bandwidth;
    unsigned i;

    stime = glutGet(GLUT_ELAPSED_TIME);

    for (i = 0; i < ctx.vid_count; ++i) {
        vid_upload_start(&ctx.vids[i]);
        bytes += ctx.vids[i].size;
    }
    while (net->packet) {
        vid_upload_packet(&ctx.vids[net->packet->vidid], net->packet);
        bytes += net->packet->w * net->packet->h * 4;
        net->packet = net->packet->next;
    }
    for (i = 0; i < ctx.vid_count; ++i) {
        vid_upload_stop(&ctx.vids[i]);
    }

    etime = glutGet(GLUT_ELAPSED_TIME);
    dtime = (etime - stime) / 1000.0;
    bandwidth = (bytes / dtime) / (1024 * 1024);

    printf("bandwidth %ldGB/s|%ldMB/s time %f\r", bandwidth >> 10, bandwidth, dtime);

    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
}

static void network_generate_fake_packet(struct net *net)
{
    const unsigned w = 64, h = 64;
    static unsigned x = 0, y = 0, color = 100;

    struct packet *pkt = (void *)&net->buffer[net->offset];
    unsigned i, j;

    net->packet = pkt;
    for (i = 0; i < 512; ++i) {
        net->offset += sizeof(*pkt) + (w*h*4);
        if (net->offset >= net->size) {
            net->offset = 0;
            break;
        }

        pkt->next = NULL;
        pkt->w = w;
        pkt->h = h;
        pkt->x = x;
        pkt->y = y;
        pkt->ptr = (void *)&pkt[1];
        for (j = 0; j < (w * h); ++j) {
            pkt->ptr[j] = color++;
        }
        x += w;
        y += h;
        if (x >= ctx.vids[0].width) x = 0;
        if (y >= ctx.vids[0].height) y = 0;
        pkt->next = (void *)&net->buffer[net->offset];
        pkt = pkt->next;
    }

    glutPostRedisplay();
}

static int ctx_vid_add(unsigned w, unsigned h, unsigned x, unsigned y)
{
    if (ctx.vid_count >= MAX_VIDS) {
        return -1;
    }

    vid_init(&ctx.vids[ctx.vid_count++], w, h, x, y);
    return 0;
}

static void glut_reshape(int width, int height)
{
    ctx.win_width = width;
    ctx.win_height = height;

    glViewport(0, 0, width, height);
}

static void glut_key(unsigned char key, int x, int y)
{
    switch (key)
    {
    case 27:
        /* Exit on escape */
        exit(1);
        break;
    default:
        return;
    }

    glutPostRedisplay();
}

static void glut_mouse(int button, int state, int x, int y)
{
}

static void glut_draw(void)
{
    unsigned i;

    /* FIXME: compute something like framerate to avoid doing too much work */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    network_process_packet(&ctx.net);

    for (i = 0; i < ctx.vid_count; ++i) {
        vid_draw(&ctx.vids[i]);
    }

    glFlush();
    glutSwapBuffers();

    network_generate_fake_packet(&ctx.net);
}

static void glut_idle(void)
{
}

int main(int argc, char **argv)
{
    int ret;

    glutInit(&argc, argv);
    glutInitContextVersion(4, 0);
    glutInitWindowSize(800, 600);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);

    glutCreateWindow("grospaquet");

    ret = glewInit();
    if (ret != GLEW_OK) {
        return ret;
    }
    if (!GLEW_VERSION_4_0) {
        printf("Need GL 4.0 at least !\n");
        return -1;
    }

    glut_reshape(800, 600);

    network_init(&ctx.net, 64 << 20);
    ctx_vid_add(320, 240, 0, 0);

    glutReshapeFunc(glut_reshape);
    glutKeyboardFunc(glut_key);
    glutDisplayFunc(glut_draw);
    glutMouseFunc(glut_mouse);
    glutIdleFunc(glut_idle);

    glutMainLoop();

    return 0;
}
